using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;


namespace Blazedust.CUAM
{
    public class Helpers
    {
		public static void ShowMainHelpText()
        {
			SuperController.singleton.ClearMessages();
			SuperController.LogMessage("");
			SuperController.LogMessage("...Scroll down to read everything...");
			SuperController.LogMessage("");
			SuperController.LogMessage("Help/Instructions");
			SuperController.LogMessage("CUAManager by Blazedust (v1.4)");
			SuperController.LogMessage("");
			SuperController.LogMessage("* This plugin manage and automatically loads character details such as hair, nails and accessories (earrings, wings, horns, tails, glasses, etc...) whether it being clothing items or linked atoms such as CustomUnityAssets when you change look or change clothing presets.");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("Features:");
			SuperController.LogMessage("* Save and Load character detail presets just like clothing presets.");
			SuperController.LogMessage("");
			SuperController.LogMessage("* Auto load character detail presets based on an active morph and personUid on the person. This allows automatically to load presets (hair, nails or other character details) automatically when changing look.");
			SuperController.LogMessage("");
			SuperController.LogMessage("* Adjust position and rotation for a linked atom asset relative to the linked body controller for a person. Now you can align and center linked atoms perfectly.");
			SuperController.LogMessage("");
			SuperController.LogMessage("* Easily add blank assets linked to a predefined rigidbody on a person, including head, neck, chest, abdomen, hip, pelvis and many more.");
			SuperController.LogMessage("");
			SuperController.LogMessage("* Supports plugin scripts attached to the linked atoms when saving the preset so it will be loaded back up again. Perfect if you have ParentHoldLink script on a CustomUnityAsset.");
			SuperController.LogMessage("");
			SuperController.LogMessage("* Can handle different atom types, not just CustomUnityAsset type atoms (additional atom types to handle can be added in the options).");
			SuperController.LogMessage("");
			SuperController.LogMessage("* Data driven, all presets and options are saved to the plugin data folder (Custom/Saves/Blazedust/CUAManager/presets/).");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("Instructions:");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.0 Selecting Active Person:");
			SuperController.LogMessage("This plugin only operates on the current selected person. This plugin will select whatever person you select in the scene or choose from the 'Active Person' selector. As you edit a character and load a look you probably already have that person selected.");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.1 Saving presets:");
			SuperController.LogMessage("You can save two different type of character detail presets, one containing clothing items and another one containing linked atoms (such as linked CustomUnityAssets).");
			SuperController.LogMessage("First type in a unique name for your preset and then press the save preset button.");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.1a Clothing preset: To save a clothing preset press the 'Save Clothing Preset: Save preset' button. All enabled clothing items for the active person will be included in the preset. Make sure to remove any clothing items that's not part of the character look (only keep hair, nails enabled etc...)");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.1b Linked atom preset: To save a linked atom preset press the 'Save Linked Atom Preset: Save preset' button. All linked atoms that this plugin is configured to handled will be included in the preset. So make sure you have the presets aligned correctly.");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.2 Loading presets:");
			SuperController.LogMessage("Select preset to load in the preset selector and press either the Replace or Append button. Replace will remove any existing presets loaded earlier for the active person while append will keep existing presets intact.");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.3 Auto load triggers:");
			SuperController.LogMessage("The strength of this plugin is that it can load all these character detail presets automatically as you change looks, no need to access this plugin interface afterwards. This is done by setting up auto load triggers.");
			SuperController.LogMessage("An auto load trigger will activate based on a morph on the character. The idea is that a unique look will have a unique morph.");
			SuperController.LogMessage("");
			SuperController.LogMessage("If you don't have a unique morph for a look, you can use an empty morph together with your look. Plugins such as MorphMerger can be used to create unique morphs that you can use with your look and auto load triggers.");
			SuperController.LogMessage("Use the *ANY* morph to make the trigger always active on all looks, if you have character details you always want to apply to all your looks, such as an iris reflections or similar.");
			SuperController.LogMessage("Note: Using the *ANY* morph together with a personUid is not guaranteed to work as changing looks can cause the personUid to change depending on the scene contents.");
			SuperController.LogMessage("");
            SuperController.LogMessage("* To create an auto load trigger: Scroll to the bottom. Select a unique trigger morph for your look and select a preset and press the 'Add Trigger' button. You can set up multiple triggers for the same trigger morph.");
			SuperController.LogMessage("");
			SuperController.LogMessage("Next time you load the look all presets set up with auto load triggers will be loaded automatically, whether it being linked CustomUnityAssets or clothing items.");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.4 Press any yellow button/label for additional help.");
			SuperController.LogMessage("");
			SuperController.LogMessage("");
			SuperController.LogMessage("* 1.5 If you want to automatically start this plugin when you start VaM check out SessionPluginBooter by Blazedust.");
			SuperController.LogMessage("");
			SuperController.LogMessage("...Scroll to top to read everything...");
			SuperController.singleton.OpenMessageLogPanel();
		}
		
		/// <summary>
		/// Helper to just return first best Atom of a specific prioritized type. 
		/// If none of the types are found any atom will be returned in the scene.
		/// </summary>
		public static Atom GetFirstAtomOfType(string[] prioAtomTypes)
		{
			Atom[] prioAtoms = new Atom[prioAtomTypes.Length + 1];
			foreach (Atom a in SuperController.singleton.GetAtoms())
			{
				bool isMatchingType = false;
				for (int i = 0; i < prioAtomTypes.Length; i++)
				{
					isMatchingType = (a.type == prioAtomTypes[i]);
					if (isMatchingType && prioAtoms[i] == null)
					{
						prioAtoms[i] = a;
					}
				}
				if (!isMatchingType && prioAtoms[prioAtoms.Length - 1] == null)
				{
					prioAtoms[prioAtoms.Length - 1] = a;
				}
			}
			return prioAtoms.Where((x) => { return x != null; }).FirstOrDefault();
		}

		public static List<string> GetLinkedAtomUids(Atom atom, Func<string, bool> atomTypeIncluded)
		{
			List<string> result = new List<string>();
			if (atom != null && !atom.destroyed)
			{
				result.AddRange(Helpers.GetChildAtoms(atom, atomTypeIncluded).Select((x) => { return x.uid; }).ToList());
				result.Sort();
			}
			return result;
		}

		public static List<string> GetRootAtomUids()
		{
			List<string> result = SuperController.singleton.GetAtoms()
			.Where((atom) => {
				return (atom.type == "Person");
			}).Select((atom) => {
				return atom.uid;
			}).ToList();
			result.Sort();
			return result;
		}

		/// <summary>
		/// Gets the DAZCharacterSelector from the specified person atom.
		/// </summary>
		public static DAZCharacterSelector GetDAZCharacterSelector(Atom rootAtom)
		{
			if (rootAtom != null && !rootAtom.destroyed && rootAtom.type == "Person")
			{
				DAZCharacterSelector geometry = rootAtom.GetStorableByID("geometry") as DAZCharacterSelector;
				return geometry;
			}
			return null;
		}

		/// <summary>
		/// Returns active clothing items for the specified person atom.
		/// </summary>
		public static IEnumerable<DAZClothingItem> GetClothingItems(Atom rootAtom)
		{
			return ClothingItems(rootAtom, (x) => { return x.active; });
		}

		/// <summary>
        /// Returns all clothing items for the specified person atom in a dictionary(uid, DAZClothingItem)
        /// </summary>
		public static Dictionary<string, DAZClothingItem> GetAllClothingItems(Atom rootAtom)
		{
			Dictionary<string, DAZClothingItem> result = new Dictionary<string, DAZClothingItem>();
			foreach (DAZClothingItem clothing in ClothingItems(rootAtom, (x) => { return true; }))
            {
				result[clothing.uid] = clothing;
			}
			return result; 
		}

		/// <summary>
		/// Enumerate clothing items for a person using a predicate.
		/// </summary>
		public static IEnumerable<DAZClothingItem> ClothingItems(Atom rootAtom, Predicate<DAZClothingItem> predicate)
        {
			DAZCharacterSelector geometry = GetDAZCharacterSelector(rootAtom);
			if (geometry != null)
			{
				foreach (DAZClothingItem clothing in geometry.clothingItems)
				{
					if (clothing != null && predicate(clothing))
					{
						yield return clothing;
					}
				}
			}	
		}

		/// <summary>
        /// Helper to remove all clothing items from the specified person.
        /// </summary>
		public static void RemoveAllClothingItems(Atom rootAtom)
        {
			DAZCharacterSelector geometry = Helpers.GetDAZCharacterSelector(rootAtom);
			if (geometry != null)
            {
				foreach(DAZClothingItem clothingItem in GetClothingItems(rootAtom))
                {
					geometry.SetActiveClothingItem(clothingItem, false);
				}
			}
		}

		/// <summary>
		/// Creates a JSON class over a clothing item.
		/// </summary>
		public static JSONClass CreateJSONClothingClass(DAZClothingItem clothing)
        {
			JSONClass jc = new JSONClass();
			System.Text.StringBuilder sb = new System.Text.StringBuilder();
			sb.Append("{");
			sb.Append(" \"clothingUid\": \"" + clothing.uid + "\",");
			sb.Append(" \"storables\": [");
			for (int i = 0; i < clothing.transform.childCount; i++)
			{
				Transform instance = clothing.transform.GetChild(0);
				JSONStorable[] componentsInChildren = instance.GetComponentsInChildren<JSONStorable>(includeInactive: true);
				foreach (JSONStorable js in componentsInChildren)
				{
					JSONClass json = js.GetJSON();
					sb.Append(json.ToString());
				}
			}
			sb.Append(" ]");
			sb.Append("}");
			return JSONNode.Parse(sb.ToString()).AsObject;
		}

		/// <summary>
		/// Creates a JSONClass representation of the atom
		/// </summary>
		public static JSONClass CreateJSONClass(Atom rootAtom, Atom atom)
		{

			string linkTo = "";
			Vector3 linkPos = new Vector3(0f, 0f, 0f);
			Vector3 linkRot = new Vector3(0f, 0f, 0f);
			LinkToInfo link = Helpers.GetLinkToInfo(atom);
			if (link != null)
			{
				// Find relevant local pos/rot information
				linkTo = link.linkToRB;
				Rigidbody rbLink = null;
				foreach (Rigidbody rb in rootAtom.linkableRigidbodies)
				{
					if (link.linkToRB == rb.name)
					{
						rbLink = rb;
						break;
					}
				}
				if (rbLink != null)
				{
					linkTo = rbLink.name;
					linkPos = rbLink.transform.InverseTransformPoint(atom.childAtomContainer.position);
					linkRot = (Quaternion.Inverse(atom.childAtomContainer.transform.rotation) * rbLink.transform.rotation).eulerAngles;
				}
			}


			JSONClass jc = new JSONClass();
			System.Text.StringBuilder sb = new System.Text.StringBuilder();
			sb.Append("{");
			sb.Append(" \"cua\": \"" + atom.uid + "\","); // cua is actually id, but kept for backwards compatability with older preset versions! :)
			sb.Append(" \"atoms\": [");
			sb.Append("  {");
			sb.Append("   \"type\": \"" + atom.type + "\",");
			sb.Append("   \"setUnlistedParamsToDefault\": \"" + "true" + "\",");
			sb.Append("   \"collisionEnabled\": \"" + (atom.collisionEnabledJSON.val ? "true" : "false") + "\",");
			sb.Append("   \"storables\": [");
			foreach (string s in atom.GetStorableIDs())
			{
				sb.Append(atom.GetStorableByID(s).GetJSON(forceStore: true).ToString());
			}
			sb.Append("     {");
			sb.Append("      \"id\": \"" + "cuabLinkTo" + "\",");
			sb.Append("      \"linkTo\": \"" + linkTo + "\",");
			sb.Append("      \"posx\": \"" + string.Format("{0:0.000000}", linkPos.x) + "\",");
			sb.Append("      \"posy\": \"" + string.Format("{0:0.000000}", linkPos.y) + "\",");
			sb.Append("      \"posz\": \"" + string.Format("{0:0.000000}", linkPos.z) + "\",");
			sb.Append("      \"rotx\": \"" + string.Format("{0:0.00000}", linkRot.x) + "\",");
			sb.Append("      \"roty\": \"" + string.Format("{0:0.00000}", linkRot.y) + "\",");
			sb.Append("      \"rotz\": \"" + string.Format("{0:0.00000}", linkRot.z) + "\"");
			sb.Append("     }");
			sb.Append("   ]");
			sb.Append("  }");
			sb.Append(" ]");
			sb.Append("}");

			return JSONNode.Parse(sb.ToString()).AsObject;
		}

		/// <summary>
		/// Gets a list over key-value defining scripts in the PluginManager for the atom. Returns null if no scripts available.
		/// </summary>
		public static List<Pair> GetAtomScripts(Atom atom)
		{
			JSONStorable store = atom.GetStorableByID("PluginManager");
			if (store != null)
			{
				JSONNode jn = store.GetJSON(true, true, true)["plugins"];
				if (jn != null && jn.Count > 0)
				{
					List<Pair> result = new List<Pair>();
					JSONClass jc = jn.AsObject;
					foreach (string key in jc.Keys)
					{
						result.Add(new Pair(key, jc[key].Value));
					}
					return result;
				}
			}
			return null;
		}

		/// <summary>
		/// Gets all child (linked) atoms of a specific type for the root atom.
		/// </summary>
		public static IEnumerable<Atom> GetChildAtoms(Atom rootAtom, Func<string, bool> atomTypeIncluded)
		{
			string parentuid = rootAtom.uid;
			foreach (Atom atom in SuperController.singleton.GetAtoms())
			{
				if (atom == null || atom.uid == parentuid || !atomTypeIncluded(atom.type))
					continue;

				LinkToInfo link = GetLinkToInfo(atom);
				if (link != null && link.linkToAtomUid == parentuid)
				{
					yield return atom;
				}
			}
		}

		public static List<string> GetSavePresetNameSuggestions(Atom rootAtom, Func<string, bool> atomTypeIncluded, params string[] predefinedSuggestions)
		{
			List<string> result = new List<string>();
			if (rootAtom != null && !rootAtom.destroyed)
			{
				// Add selected preset name to allow for easy overwrite possibility
				if (predefinedSuggestions != null && predefinedSuggestions.Length > 0)
				{
					result.AddRange(predefinedSuggestions.Where((x) => { return (x != null && x != ""); }));
				}

				// Child CustomUnityAsset atom uids prefix to ignore. They are automatically tested against "", "#", "_" separators.
				string[] saveNamePrefixCuts = new string[2] { "CUA", "CustomUnityAsset" };

				Func<Atom, string[]> cuaAtomUidSelector = new Func<Atom, string[]>((x) => {
					foreach (string prefix in saveNamePrefixCuts)
					{
						foreach (string prefixVariant in new string[3] { prefix + "#", prefix + "_", prefix }) // prefix + " "  <-- Depends on your taste...
						{
							if (x.uid.StartsWith(prefixVariant) && x.uid.Length > prefixVariant.Length)
							{
								string postfixName = x.uid.Substring(prefixVariant.Length).Trim();
								// To avoid names like "2" or "3" we just try to parse the postfix here. If conversion fails it's not an auto-generated post-number-sequence from VaM and we include that as a valid name.
								int nr = 0;
								if (!int.TryParse(postfixName, out nr))
								{
									// additionally check for "recursive" "#2" or "2#" stuff.
									bool postfixNonsense = (postfixName.Length == 2 && (postfixName.EndsWith("#") || postfixName.StartsWith("#")));
									if (!postfixNonsense)
									{
										return new string[2] { x.uid.Trim(), postfixName };
									}
								}
							}
						}
					}
					return new string[1] { x.uid.Trim() };
				});

				// Add all child-CustomUnityAssets uids as selectable names. This makes sense when you have only one asset you want to save as a preset.
				List<string> ids = Helpers.GetChildAtoms(rootAtom, atomTypeIncluded).SelectMany((x) => {
					return cuaAtomUidSelector(x);
				}).ToList();

				ids.Sort();
				result.AddRange(ids);

				// add Person's uid on top as a preset name suggestion as long as its NOT "Person".
				if (!rootAtom.uid.StartsWith("Person"))
				{
					result.Insert(0, rootAtom.uid);
				}

				// Add clothing suggestions next, keep them in their own separate group
				foreach (DAZClothingItem clothingItem in GetClothingItems(rootAtom))
				{
					result.Add(clothingItem.displayName);
				}

				// avoid duplicates
				result = result.Distinct().ToList();
			}

			return result;
		}

		// THIS FUNCTION IS OBSOLETE, USED DURING TESTING AND DEVELOPMENT OF PLUGIN
		/*public static void ReadChildAtoms(Atom parentAtom, Func<string, bool> atomTypeIncluded)
		{
			string parentuid = parentAtom.uid;
			foreach (Atom atom in SuperController.singleton.GetAtoms())
			{
				if (atom == null || atom.uid == parentuid || !atomTypeIncluded(atom.type))
					continue;

				LinkToInfo link = GetLinkToInfo(atom);
				if (link != null && link.linkToAtomUid == parentuid)
				{
					Rigidbody rbLink = null;
					foreach (Rigidbody rb in parentAtom.linkableRigidbodies)
					{
						if (link.linkToRB == rb.name)
						{
							rbLink = rb;
							break;
						}
					}
					if (rbLink != null)
					{
						// Figure out local space
						Vector3 pos = rbLink.transform.InverseTransformPoint(atom.childAtomContainer.position);
						Vector3 rot = (Quaternion.Inverse(atom.childAtomContainer.transform.rotation) * rbLink.transform.rotation).eulerAngles;
						// SuperController.LogMessage(atom.uid + " is linked to " + link.linkToAtomUid + " + " + link.linkToRB + " with:");
						// SuperController.LogMessage("local pos: " + pos.ToString());
						// SuperController.LogMessage("local rot: " + rot.ToString());

						string btnLine = "new LinkableAsset(\"" + rbLink.name + " (" + atom.uid + ")\", \"" + rbLink.name + "\", \"CUA_" + rbLink.name + "\", " + string.Format(" new Vector3({0:0.000}f, {1:0.000}f, {2:0.000}f)", pos.x, pos.y, pos.z) + ", " + string.Format(" new Vector3({0:0.0}f, {1:0.0}f, {2:0.0}f)", rot.x, rot.y, rot.z) + "\"),";

						SuperController.LogMessage("--");
						SuperController.LogMessage(btnLine);
					}
				}
			}
			SuperController.LogMessage("--");
		}*/

		public static LinkToInfo GetLinkToInfo(Atom atom)
		{
			// Can read link info from GetJSON
			// JSONNode to read is constructed like this: jSON["linkTo"] = linkToAtomUID + ":" + _linkToRB.name;
			if (atom.mainController != null)
			{
				JSONNode node = atom.mainController.GetJSON()["linkTo"];
				if (node != null)
				{
					string linkTo = node.Value;
					if (linkTo != null && linkTo != "")
					{
						int splitIndex = linkTo.LastIndexOf(":");
						if (splitIndex > 0)
						{
							return new LinkToInfo(linkTo.Substring(0, splitIndex), linkTo.Substring(splitIndex + 1));
						}
					}
				}
			}

			return null;
		}

		public static AssetDataInfo GetAtomAssetInfo(Atom atom)
		{
			AssetDataInfo data = new AssetDataInfo();

			JSONStorable asset = atom.GetStorableByID("asset");
			if (asset == null)
				return data;

			JSONClass assetjc = asset.GetJSON(forceStore: true);
			if (assetjc == null)
				return data;

			// SuperController.LogMessage("importLightmaps: " + assetjc["importLightmaps"].Value);
			// SuperController.LogMessage("importLightProbes: " + assetjc["importLightProbes"].Value);
			// SuperController.LogMessage("registerCanvases: " + assetjc["registerCanvases"].Value);
			// SuperController.LogMessage("showCanvases: " + assetjc["showCanvases"].Value);
			// SuperController.LogMessage("loadDll: " + assetjc["loadDll"].Value);
			// SuperController.LogMessage("assetName: " + assetjc["assetName"].Value);
			// SuperController.LogMessage("assetUrl: " + assetjc["assetUrl"].Value);
			// SuperController.LogMessage("assetDllUrl: " + assetjc["assetDllUrl"].Value);

			data.assetName = assetjc["assetName"].Value;
			data.assetUrl = assetjc["assetUrl"].Value;
			data.assetDllUrl = assetjc["assetDllUrl"].Value;
			data.assetBundleName = Path.GetFileNameAndExt(data.assetUrl);
			return data;
		}

		/// <summary>
		/// Get preset meta info from preset file
		/// </summary>
		public static List<PresetItemMetaInfo> GetPresetMetaInfo(string file, Func<string, bool> atomTypeIncluded)
		{
			try
			{
				JSONNode jn = SuperController.singleton.LoadJSON(file);
				JSONClass jc = jn.AsObject;

				List<PresetItemMetaInfo> result = new List<PresetItemMetaInfo>();

				if (jn["items"] != null)
				{
					// linkedAtom preset
					foreach (JSONNode itemNode in jn["items"].AsArray.Childs)
					{
						PresetItemMetaInfo item = GetPresetMetaInfoFromNode(itemNode, atomTypeIncluded);
						if (item != null)
						{
							result.Add(item);
						}
					}
				}

				if (jn["clothing"] != null)
				{
					// Clothing preset
					foreach (JSONNode itemNode in jn["clothing"].AsArray.Childs)
					{
						PresetItemMetaInfo item = GetPresetClothingMetaInfoFromNode(itemNode);
						if (item != null)
						{
							result.Add(item);
						}
					}
				}

				return result;
			}
			catch (Exception e)
			{
				SuperController.LogError("Error getting meta info from preset file: " + file);
				SuperController.LogError(e.ToString());
			}
			return null;
		}

		public static PresetItemMetaInfo GetPresetClothingMetaInfoFromNode(JSONNode cuaNode)
        {
			string clothingId = cuaNode["clothingUid"].Value;
			PresetItemMetaInfo item = new PresetItemMetaInfo();
			item.isClothingPreset = true;
			item.clothingUid = clothingId;
			return item;
		}

		public static PresetItemMetaInfo GetPresetMetaInfoFromNode(JSONNode cuaNode, Func<string, bool> atomTypeIncluded)
		{
			// Log("Creating " + cuaNode["cua"].Value + " from preset");
			string atomId = cuaNode["cua"].Value;
			JSONNode jnAtom = cuaNode["atoms"].AsArray.Childs.FirstOrDefault((x) => { return atomTypeIncluded(x["type"].Value); });
			if (jnAtom == null)
			{
				return null;
			}

			JSONNode jnCuabLinkTo = jnAtom["storables"].AsArray.Childs.Where((x) => { return x["id"].Value == "cuabLinkTo"; }).FirstOrDefault();
			if (jnCuabLinkTo == null)
			{
				return null;
			}

			string atomType = cuaNode["atoms"].AsArray.Childs.FirstOrDefault()["type"].Value;

			string linkRBName = jnCuabLinkTo["linkTo"];
			Vector3 localPos = new Vector3(
					float.Parse(jnCuabLinkTo["posx"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["posy"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["posz"].Value, System.Globalization.CultureInfo.InvariantCulture));

			Vector3 localRot = new Vector3(
					float.Parse(jnCuabLinkTo["rotx"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["roty"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["rotz"].Value, System.Globalization.CultureInfo.InvariantCulture));

			string assetName = "";
			string assetBundleName = "";
			if (atomType == "CustomUnityAsset")
            {
				JSONNode assetNode = jnAtom["storables"].AsArray.Childs.Where((x) => { return x["id"].Value == "asset"; }).FirstOrDefault();
				if (assetNode != null && assetNode["assetName"] != null && assetNode["assetName"].Value != null && assetNode["assetUrl"] != null && assetNode["assetUrl"].Value != null)
                {
					assetName = assetNode["assetName"].Value;
					assetBundleName = Path.GetFileNameAndExt(assetNode["assetUrl"].Value);
				}
			}

			PresetItemMetaInfo item = new PresetItemMetaInfo();
			item.isLinkedAssetPreset = true;
			item.atomType = atomType;
			item.linkRBName = linkRBName;
			item.atomId = atomId;
			item.localPosition = localPos;
			item.localRotation = localRot;
			item.assetName = assetName;
			item.assetBundleName = assetBundleName;
			return item;
		}

		/// <summary>
		/// Removes a matching scene preset (presetMetaInfoItems) from a parent atom.
		/// </summary>
		public static void RemoveScenePreset(Atom parentAtom, List<PresetItemMetaInfo> presetMetaInfoItemsToRemove, Func<string, bool> atomTypeIncluded)
        {
			//SuperController.LogMessage("Checking matching " + presetMetaInfoItems.Count.ToString() + " presets");
			List<Atom> matchingAtoms = new List<Atom>();
			HashSet<PresetItemMetaInfo> matchingPresetMetaInfos = new HashSet<PresetItemMetaInfo>();
			foreach(Atom atom in GetChildAtoms(parentAtom, atomTypeIncluded))
            {
				if (atom.destroyed)
					continue;

				LinkToInfo linkToInfo = GetLinkToInfo(atom);
				if (linkToInfo == null)
					continue;

				foreach (PresetItemMetaInfo metaInfo in presetMetaInfoItemsToRemove)
                {
					if (atom.type == metaInfo.atomType && metaInfo.isLinkedAssetPreset && (atom.uid.StartsWith(metaInfo.atomId) || atom.type == "CustomUnityAsset") && linkToInfo.linkToRB == metaInfo.linkRBName)
                    {
						if (atom.type == "CustomUnityAsset")
                        {
							// asset bundle data must match for CustomUnityAssets
							AssetDataInfo assetInfo = GetAtomAssetInfo(atom);
							if (metaInfo.assetBundleName == assetInfo.assetBundleName && metaInfo.assetName == assetInfo.assetName)
                            {
								// MATCH!
								matchingAtoms.Add(atom);
								matchingPresetMetaInfos.Add(metaInfo);
								//SuperController.LogMessage("Matching CUA meta info: " + atom.uid);
							}
						} 
						else
                        {
							// MATCH
							matchingAtoms.Add(atom);
							matchingPresetMetaInfos.Add(metaInfo);
							//SuperController.LogMessage("Matching meta info: " + atom.uid);
						}
					}
				}
			}

			// Remove matching atoms if all preset meta infos was matched.
			if (matchingPresetMetaInfos.Count == presetMetaInfoItemsToRemove.Count)
            {
				matchingAtoms.ForEach((x) => {
					try
					{
						x.Remove();
					} 
					catch (Exception e)
                    {
						SuperController.LogError("Removing atom failed with:");
						SuperController.LogError(e.ToString());
                    }
				});
			}
		}
	}

	public class UpdateAtomPostCreation
	{
		public Atom apAtom;
		public FreeControllerV3.PositionState posState;
		public FreeControllerV3.RotationState rotState;
		public bool collisionEnabled;
		public bool physicsEnabled;
		public int delayFrames;
	}

	public class PresetItemMetaInfo
	{
		public bool isClothingPreset;
		public bool isLinkedAssetPreset;

		// For Clothing presets
		public string clothingUid;

		public string atomType;
		public string linkRBName;
		public string atomId;
		public Vector3 localPosition;
		public Vector3 localRotation;
		
		// For CustomUnityAssets
		public string assetName;
		// For CustomUnityAssets
		public string assetBundleName;

		public PresetItemMetaInfo()
		{
			isClothingPreset = false;
			isLinkedAssetPreset = false;
			clothingUid = "";
			atomType = "";
			linkRBName = "";
			atomId = "";
			localPosition = Vector3.zero;
			localRotation = Vector3.zero;
			assetName = "";
			assetBundleName = "";
		}
	}

	public class AssetDataInfo
	{
		public string assetName;
		public string assetBundleName;
		public string assetUrl;
		public string assetDllUrl;
	}

	/// <summary>
	/// Represents a generic linkable asset (hardcoded) available in this plugin.
	/// </summary>
	public class LinkableAsset
	{
		public string name;
		public string linkRBName;
		public string atomPrefix;
		public Vector3 localPosition;
		public Vector3 localRotation;
		public bool guiRightSide;

		public LinkableAsset(string btnName, string linkRB, string prefix, Vector3 localPos, Vector3 localRot, bool guiRight = false)
		{
			name = btnName;
			linkRBName = linkRB;
			atomPrefix = prefix;
			localPosition = localPos;
			localRotation = localRot;
			guiRightSide = guiRight;
		}
	}

	public class LinkToInfo
	{
		/// This is the parent atom uid it's linked to
		public string linkToAtomUid = "";
		/// This is the sub control it's linked to (chest/leg/hand/ etc...)
		public string linkToRB = "";

		public LinkToInfo(string atomUid, string rb)
		{
			linkToAtomUid = atomUid;
			linkToRB = rb;
		}
	}

	public class Pair
	{
		public string key;
		public string value;

		public Pair()
		{
			key = "";
			value = "";
		}
		public Pair(string k, string v)
		{
			key = k;
			value = v;
		}

		public override string ToString()
		{
			return string.Format("key: {0}, value: {1}", key, value);
		}
	}


}