using System;
using UnityEngine;
using UnityEngine.EventSystems;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{
	public class UILoading
	{
		CustomUnityAssetManager core;

		public int forcedLoadingPostDelayUpdates = 0;
		public int loadingCounter = 0;

		public UILoading(CustomUnityAssetManager script)
        {
			core = script;
		}

		public void Destroy()
		{
			core = null;
		}

		public bool LoadingInProgress(int maxForcedLoadingPostDelayUpdates = 0)
		{
			return (SuperController.singleton.isLoading || loadingCounter > 0 || forcedLoadingPostDelayUpdates > maxForcedLoadingPostDelayUpdates);
		}

		public void LoadingStart(string msg)
		{
			loadingCounter++;
			// Force post-delay waiting
			forcedLoadingPostDelayUpdates = Math.Max(110, forcedLoadingPostDelayUpdates + 110);
			SetLoadingText(msg);
		}		

		public void LoadingEnd()
		{
			loadingCounter--;
			if (loadingCounter <= 0)
			{
				SetLoadingText("Post waiting...");
				forcedLoadingPostDelayUpdates = Math.Max(1, forcedLoadingPostDelayUpdates);
			}
		}

		public void SetLoadingText(string msg)
		{
			core.infoTextBtn.buttonText.text = msg;
			if (SuperController.singleton.loadingUIAlt != null)
			{
				SuperController.singleton.loadingUIAlt.gameObject.SetActive((msg != ""));
				if (SuperController.singleton.loadingTextStatusAlt != null)
				{
					SuperController.singleton.loadingTextStatusAlt.text = msg;
				}
			}
		}

		public bool PostDelay()
        {
			if (forcedLoadingPostDelayUpdates > 0)
			{
				if (loadingCounter <= 0)
				{
					forcedLoadingPostDelayUpdates--;
					if (forcedLoadingPostDelayUpdates <= 0)
					{
						SetLoadingText("");
					}
				}
				// Post delay
				return true;
			}
			return false;
		}

	}
}